from threading import Lock
from typing import Any, Dict, Optional, Set, List
from dataclasses import dataclass, field
from collections.abc import Iterable
from maspy.utils import utils
from maspy.error import InvalidPerceptError

DEFAULT_GROUP = "default"

@dataclass
class Percept:
    key: str
    args: tuple = field(default_factory=tuple)
    group: str = DEFAULT_GROUP # Percept Type (still unsure)
    
    def __hash__(self) -> int:
        args_hashable = []
        for arg in self.args:
            args_hashable.append(arg)

        args_hashable = tuple(args_hashable)

        return hash((self.key, args_hashable, self.group))

class EnvironmentMultiton(type):
    _instances: Dict[str, "Environment"] = {}
    _lock: Lock = Lock()

    def __call__(cls, __my_name="env"):
        with cls._lock:
            if __my_name not in cls._instances:
                instance = super().__call__(__my_name)
                cls._instances[__my_name] = instance
        return cls._instances[__my_name]

class Environment(metaclass=EnvironmentMultiton):
    def __init__(self, env_name: str):
        self.full_log = False
        self._my_name = env_name
        self.agent_list = {}
        self._agents = {}
        self._name = f"Environment:{self._my_name}"
        self._percepts = dict()
    
    def print(self,*args, **kwargs):
        return print(f"{self._name}>",*args,**kwargs)
    
    def perception(self):
        return self._percepts

    @property
    def print_percepts(self):
        percepts = ""
        for group_keys in self._percepts.values():
            for percept_set in group_keys.values():
                for percept in percept_set:
                    percepts += f"{percept}\n"
        print(f"{percepts}\r")
        
    def add_agents(self, agents):
        try:
            for agent in agents:
                self._add_agent(agent)
        except TypeError:
            self._add_agent(agents)
    
    def _add_agent(self, agent):
        if type(agent).__name__ in self.agent_list:
            if agent.my_name[0] in self.agent_list[type(agent).__name__]:
                self.agent_list[type(agent).__name__][agent.my_name[0]].update({agent.my_name})
                self._agents[agent.my_name] = agent
            else:
                self.agent_list[type(agent).__name__].update({agent.my_name[0] : {agent.my_name}})
                self._agents[agent.my_name] = agent
        else:
            self.agent_list[type(agent).__name__] = {agent.my_name[0] : {agent.my_name}}
            self._agents[agent.my_name] = agent
        
        self.print(f'Connecting agent {type(agent).__name__}:{agent.my_name}')
    
    def create_percept(self, 
            percept: Iterable[Percept] | Percept
        ):
        percept_dict = self._clean(percept)
            
        self._percepts = utils.merge_dicts(self._percepts,percept_dict)
        self.print(f"Creating percept {percept_dict}") if self.full_log else ...

    def change_percept(self, key: str, old_args: Any = tuple(), new_args: Any = tuple(), group: Optional[str] = DEFAULT_GROUP):
        if type(old_args) is not tuple: old_args = (old_args,) 
        if type(new_args) is not tuple: new_args = (new_args,) 
        for percept in self._percepts[group][key]:
            if percept == Percept(key,old_args,group): 
                self.print(f"Updating {percept} to", end="")
                percept.args = new_args
                print(f" {percept}") 
                break   
            
    def _percept_exists(self, key, args, group=DEFAULT_GROUP) -> bool:
        if type(args) is not tuple: args = (args,)
        return Percept(key,args,group) in self._percepts[group][key]

    def delete_percept(self, 
            key: str, 
            args: Optional[Any] = tuple(), 
            group: Optional[str] = DEFAULT_GROUP,
        ):
        if type(args) is not tuple: args = (args,) 
        self._percepts[group][key].remove(Percept(key,args,group))
              
    def _clean(
        self, percept_data: Iterable[Percept] | Percept
    ) -> Dict:
        match percept_data:
            case None:
                return dict()
            case Percept():
                return {self._my_name : {percept_data.key: {percept_data}}}
            case Iterable():
                percept_dict = dict()
                for prc_dt in percept_data:
                    if prc_dt.key in percept_dict[self._my_name]:
                        percept_dict[prc_dt.group][prc_dt.key].add(prc_dt)
                    else:
                        percept_dict[prc_dt.group].update({prc_dt.key: {prc_dt}})

                return percept_dict
            case _:
                raise TypeError(
                    f"Expected data type to have be Iterable[Percept] | Percept, recieved {type(percept_data).__name__}"
                ) 




from typing import( Generic, Optional, Sequence, Tuple, Type, TypeVar, Union )
from contextlib import closing
from time import sleep
from io import StringIO
from os import system
import numpy as np
import sys
import matplotlib.pyplot as plt

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")
T_cov = TypeVar("T_cov", covariant=True)

class learning_model(object):
    def __init__(self, map = None):
        if map is None: # default 4x4 map
            map = [[-2, 0, 0,-2],
                   [-1, 0, 0, 0],
                   [ 0,-2,-1,-1],
                   [ 0, 0, 0,-2]]
            
        self.desc = np.asarray(map, dtype=int)
        self.ori_desc = np.asarray(map, dtype=int).copy()
        
        locs = np.where(self.desc == -2)
        self.locs = [ (x,y) for x,y in zip(locs[0],locs[1]) ] 
        self.comb = 2**len(self.locs)

        self.num_rows, self.num_cols = self.desc.shape    
        self.num_states = self.num_rows * self.num_cols
        self.initial_state_distrib = np.zeros(self.num_states)
        self.max_row = self.num_rows - 1
        self.max_col = self.num_cols - 1
        self.num_actions = 4 # down, up, right, left
        
        self.P = {
            state: {action: [] for action in range(self.num_actions)}
            for state in range(self.num_states)
        }
        
        arrive_reward = 10
        for row in range(self.num_rows):
            for col in range(self.num_cols):
                state = self.encode(row, col)
                if self.desc[row,col] == 0:   # Estados iniciais devem estar fora de sujeira e obstaculos
                    self.initial_state_distrib[state] += 1
                for action in range(self.num_actions):
                    new_row, new_col, = row, col
                    reward = 0 # Recompensa para movimentar normalmente
                    if action == 0 and row < self.max_row and self.desc[row+1, col] != -1 : #  Movimenta p/ Baixo
                        new_row = min(row+1,self.max_row)
                        if self.desc[new_row,col] == -2:    # Recompensa quando alcança posicao alvo
                            reward = arrive_reward
                    elif action == 1 and row > 0 and self.desc[row-1, col] != -1: # Movimenta p/ Cima
                        new_row = max(row-1, 0)
                        if self.desc[new_row,col] == -2:    # Recompensa quando alcança posicao alvo
                            reward = arrive_reward
                    elif action == 2 and col < self.max_col and self.desc[row, col+1] != -1: # Movimenta p/ Direita
                        new_col = min(col+1,self.max_col)
                        if self.desc[row,new_col] == -2:    # Recompensa quando alcança posicao alvo
                            reward = arrive_reward
                    elif action == 3 and col > 0 and self.desc[row, col-1] != -1: # Movimenta p/ Esquerda
                        new_col = max(col-1,0) 
                        if self.desc[row,new_col] == -2:    # Recompensa quando alcança posicao alvo
                            reward = arrive_reward
                    
                    new_state = self.encode(new_row,new_col) # Novo estado apos acao
                    self.P[state][action].append( (1.0, new_state, reward) )

        self.initial_state_distrib /= self.initial_state_distrib.sum()
        self.action_space = Discrete(self.num_actions)
        self.observation_space = Discrete(self.num_states)

    # Recuperar Estado Atual de Acordo com Observacao
    def encode(self, row, col):
        return  row * self.num_cols + col

    # Recuperar Observacao de Acordo com Estado Atual
    def decode(self, i):
        out = []
        out.append(i % self.num_rows)
        i = i // self.num_rows
        out.append(i)
        return reversed(out)

    # Mascara de Possiveis Acoes
    def action_mask(self, state: int):
        mask = np.zeros(self.num_actions, dtype=np.int8)
        row, col = self.decode(state)
        if row < self.max_row and self.desc[row+1, col] != -1: 
            mask[0] = 1
        if row > 0 and self.desc[row-1, col] != -1:
            mask[1] = 1
        if col < self.max_col and self.desc[row, col+1] != -1: 
            mask[2] = 1
        if col > 0 and self.desc[row, col-1] != -1:
            mask[3] = 1
        return mask

    # Retorna o resultado de uma determinada Acao 
    def look(self, a: ActType) -> Tuple[ObsType, float, bool, dict]:
        transitions = self.P[self.s][a]
        if len(transitions) > 1:
            i = self.categorical_sample([t[0] for t in transitions])
            p, s, r = transitions[i]
        else:
            p, s, r = transitions[0]

        return (int(s), r, {"prob": p, "action_mask": self.action_mask(s)})

    # Realiza uma Acao no Estado Atual
    
    def step(self, a: ActType) -> Tuple[ObsType, float, bool, dict]:
        transitions = self.P[self.s][a]
    
        if len(transitions) > 1:
            i = self.categorical_sample([t[0] for t in transitions])
            p, s, r = transitions[i]
        else:
            p, s, r = transitions[0]

        row,col = self.decode(s)
        if self.desc[row][col] == -2:
            self.desc[row][col] = 0
        elif self.desc[row][col] == 0 and r > 0:
            r = 0
            
        self.s = s
        self.last_action = a
        return (int(s), r, {"prob": p, "action_mask": self.action_mask(s)})

    # Transforma o caminho de Estados em sequencia de Acoes
    def states_to_actions(self, path):
        path_actions = []
        x, y = self.decode(self.s)
        for state in path:
            row, col = self.decode(state)
            dif = np.subtract((x,y),(row,col))
            if dif[0] != 0:
                if dif[0] < 0:
                    path_actions.append(0) # Baixo
                else:
                    path_actions.append(1) # Cima
            elif dif[1] != 0:
                if dif[1] < 0:
                    path_actions.append(2) # Direita
                else:
                    path_actions.append(3) # Esquerda
            else:
                return None                # Error
            x = row
            y = col
        #print(f'{self.s} {path} -> {path_actions}')
        return path_actions

    def actions_to_states(self,actions):
        states = []
        for action,value in enumerate(actions):
            if value == 1:
                s, _, _, = self.look(action)
                states.append(s)
        return states
    
    # Retorna possiveis Estados vizinhos e as melhores Acoes
    def sensor(self):
        best_actions = np.zeros((self.num_actions),dtype=int)
        possible_states = []
        row,col = self.decode(self.s)
        mask = self.action_mask(self.s)
        for i in range(max(row-1,0),min(row+1,self.max_row)+1):
            for j in range(max(col-1,0),min(col+1,self.max_col)+1):
                if self.desc[i,j] >= 0 and np.abs((row+col)-(i+j)) == 1:
                    possible_states.append(self.encode(i,j))
                
                if self.desc[i,j] == -2:
                    if i > row and mask[0] == 1:
                        best_actions[0] += 1
                    elif i < row and mask[1] == 1:
                        best_actions[1] += 1
                    if j > col and mask[2] == 1:
                        best_actions[2] += 1
                    elif j < col and mask[3] == 1:
                        best_actions[3] += 1
        return possible_states, best_actions
                
    # Resetar Estado para um Possivel Incial
    def reset(self, *, seed: Optional[int] = None, options: Optional[dict] = None) -> Tuple[ObsType, dict]: 
        self.s = self.categorical_sample(self.initial_state_distrib)
        self.desc = self.ori_desc.copy()
        self.last_action = None
        
        return int(self.s), {"prob": 1.0, "action_mask": self.action_mask(self.s)}

    # Coloca o sistema em um Estado especifico
    def set_state(self, state) -> Tuple[ObsType, dict]: 
        self.s = state
        self.last_action = None
        
        return int(self.s), {"prob": 1.0, "action_mask": self.action_mask(self.s)}

    # Imprimir na tela uma representacao do cenario no estado atual
    def render(self, adj_matrix=None):
        desc = self.desc.copy().tolist()
        outfile = StringIO()

        row, col = self.decode(self.s)
    
        # Representacao do aspirador no cenario 
        desc[row][col] = 1

        end = out = '+'+ ''.join(['-' for x in range(2*len(desc)-1)]) + '+\n'
        for k,line in enumerate(desc):
            out += '|'
            for i,x in enumerate(line):
                if x == -2:
                    out += '#'
                if x == -1:
                    out += '■'
                if x == 1:
                    out += 'O'
                if x == 0:
                    state = self.encode(k,i)
                    if len(adj_matrix) > 0 and adj_matrix[state,state] > 0:
                        if adj_matrix[state,state] == 1:
                            out += '*'
                        elif adj_matrix[state,state] == 2:
                            out += '~'
                        elif adj_matrix[state,state] == 3:
                            out += 'X'
                        else:
                            out += 'ADJ_ERROR'
                    else:
                        out += ' '
                if i < len(line)-1:
                    out += ':'
            out += '|\n'
        out += end

        outfile.write(out)
        if self.last_action is not None:
            outfile.write(f"({['Baixo','Cima','Direita','Esquerda'][self.last_action]})\n")
        else:
            outfile.write("\n")

        with closing(outfile):
            return outfile.getvalue()

    def categorical_sample(self, prob_n):
        np_rand = np.random.default_rng()
        prob_n = np.asarray(prob_n)
        csprob_n = np.cumsum(prob_n)
        return np.argmax(csprob_n > np_rand.random())

class Space(Generic[T_cov]):
    def __init__(self, shape: Optional[Sequence[int]] = None, dtype: Optional[Union[Type, str, np.dtype]] = None ):
        self._shape = None if shape is None else tuple(shape)
        self.dtype = None if dtype is None else np.dtype(dtype)

class Discrete(Space[int]):
    def __init__(self, n: int, start: int = 0 ):
        assert isinstance(n, (int, np.integer))
        assert n > 0, "n (counts) have to be positive"
        assert isinstance(start, (int, np.integer))
        self.n = int(n)
        self.start = int(start)
    
    def sample(self) -> int:
        return int(self.start + np.random.randint(self.n))


    # value_function
    # rewards
    # policy
    # size
    # epsilon

    # num_population
    # num_iteration

    # max_targets

class learning:
    def __init__(self, map_size=10, num_targets=10, wall_perc=0.1,
                epsilon=0.1, num_pop=5, num_step=50, num_iter=100
                ) -> None:     
        self.model = learning_model 
        self.map_size = map_size
        self.num_targets = num_targets
        self.wall_perc = wall_perc
        self.epsilon = epsilon
        self.num_pop = num_pop
        self.num_step = num_step
        self.num_iter = num_iter
        
        self.make_graph = False
        self.adj_matrix = None
        self.methods = [self.q_learning, self.sarsa]
        self.policy = None
        
    def normalize(self, numbers):
        min_val = min(numbers)
        max_val = max(numbers)
        total = sum(numbers)
        range_value = max_val - min_val
        if range_value == 0:
            return [0 for num in numbers]
        else:
            return [(num - min_val)/total for num in numbers]

    def monte_carlo_selection(self, probabilities):
        prob_aux = self.normalize(probabilities)
    
        rand_num = np.random.uniform(0, 1)
    
        position = 0
        for i, p in enumerate(prob_aux):
            if rand_num <= p:
                position = i
                break
            else:
                rand_num -= p
    
        return position

    def shortest_path(self, adj_matrix,start_state,q_table=[],end_state=None):
        n = len(adj_matrix)
        
        seen = [False]*n
        seen[start_state] = True
        distances = [-1]*n
        distances[start_state] = 0
        parent = [-1]*n

        queue = []
        queue.append(start_state)

        closest_state = -1
        best_value = -1

        while queue:
            state = queue.pop(0)
            seen_flag = False
            for vertex, edge in enumerate(adj_matrix[state]):
                if edge != 0 and not seen[vertex] and adj_matrix[vertex,vertex] != 3:
                    seen_flag = True
                    seen[vertex] = True
                    distances[vertex] = distances[state] + 1
                    parent[vertex] = state
                    queue.append(vertex)
                    if adj_matrix[vertex,vertex] < 2:
                        try:
                            vertex_value = (max(q_table[vertex])*4)/distances[vertex]
                            if vertex_value > best_value:
                                closest_state = vertex
                                best_value = vertex_value
                        except:
                            closest_state = vertex
                            queue.clear()
        
        if closest_state != -1:
            path = []
            current = closest_state
            while current != start_state:
                path.append(current)
                current = parent[current]

            return list(reversed(path))
        else:    
            print(f'Return None for shortest path: {start_state} {state} {closest_state} ')
            print(f'Submatrix:\n{self.submatrix(adj_matrix,[state,state],1)}')
            return None

    def show_frames(self, frames):
        for i, frame in enumerate(frames):
            system('cls')
            print(frame['frame'])
            print(f"Timestep: {i + 1} of {len(frames)}")
            print(f"State: {frame['state']}")
            print(f"Action: {frame['action']}")
            print(f"Action Flag: {frame['action_flag']}")
            print(f"Best Actions: {frame['best_actions']}")
            print(f"Possible Actions: {frame['possible_actions']}")
            print(f"Reward: {frame['reward']}")
            print(f"Acumulated Reward: {frame['acumulated_reward']}")
            sys.stdout.flush()
            sleep(0.8)

    def random_map(self, size, num_targets, max_wall_perc=0.1):
        wall_num = int(max_wall_perc*size*size) + np.random.randint(int(max_wall_perc*size*size))
        wall_aux = wall_num
        dirt_num = num_targets
        dirt_aux = dirt_num

        #print(f'Making {(size,size)} map with {wall_num} walls and {dirt_num} dirts')
        total_path = []
        while len(total_path) < size*size - wall_num + 1:
            map = np.zeros([size, size])
            done = False
            while not done:
                if wall_num > 0:
                    x = np.random.randint(size-1)
                    y = np.random.randint(size-1)
                    if map[x,y] == 0:
                        map[x,y] = -1
                        wall_num -= 1

                if dirt_num > 0:
                    x = np.random.randint(size-1)
                    y = np.random.randint(size-1)
                    if map[x,y] == 0:
                        map[x,y] = -2
                        dirt_num -= 1

                if wall_num == 0 and dirt_num == 0:
                    done = True
            
            #Checking map traversability
            wall_num = wall_aux
            dirt_num = dirt_aux
            coord = np.random.randint(low=0,high=size-1,size=2)
            graph = [coord]
            total_path = []
            seen = np.zeros((size,size),dtype=int)
            while graph:
                pos = graph.pop(0)
                total_path.append(pos)
                for i in range(max(0,pos[0]-1),min(size-1,pos[0]+1)+1):
                    if map[i,pos[1]] != -1 and seen[i,pos[1]] == 0:
                        graph.append([i,pos[1]])
                        seen[i,pos[1]] = 1
                for j in range(max(0,pos[1]-1),min(size-1,pos[1]+1)+1):
                    if map[pos[0],j] != -1 and seen[pos[0],j] == 0:
                        graph.append([pos[0],j])
                        seen[pos[0],j] = 1
        try:
            return map
        except Exception as e:
            print(f"{len(total_path) < size*size-wall_num} : {e}")
            sys.exit()

    def submatrix(self, matrix,center,radius):
        x, y = center
        subm_rows = slice(max(0, x-radius), min(x+radius+1, matrix.shape[0]))
        subm_cols = slice(max(0, y-radius), min(y+radius+1, matrix.shape[1]))
        subm = np.zeros((2*radius+1, 2*radius+1), dtype=matrix.dtype)
        subm[:] = -1
        subm[radius-x+subm_rows.start:radius-x+subm_rows.stop, 
            radius-y+subm_cols.start:radius-y+subm_cols.stop] = matrix[subm_rows, subm_cols]
        return subm

    def q_learning(self, q_table, epsilon, e, num_steps=50, num_iter=500):
        alpha = 0.1
        gamma = 0.9 
        sum_acumulated_reward = []
        size = e.num_states
        # Matrix de adjacencia do grafo de exploracao  
        for idx in range(1,num_iter+1):
            adj_matrix = np.zeros((size, size), dtype=int)
            state, _ = e.reset()    # Resetando aspirador para um estado aleatorio
            acumulated_reward = 0

            adj_matrix[state,state] = 2 # Este valor indica na matrix que este estado ja foi explorado
            path_actions = []
            best_states = []
            steps = 0
            while steps < num_steps:
                action, possible_states, path_actions = self.choose_action(e, epsilon, state, q_table, adj_matrix, path_actions)
                next_state, reward, _ = e.step(action)
                acumulated_reward += reward
                
                if next_state in best_states:
                    best_states.remove(next_state)
                
                # Ajustes na matrix de adjacencia
                # Inserindo ligacoes entre estados (valor 1) e estados esplorados (valor 2)
                adj_matrix[next_state,next_state] = 2
                for p_state in possible_states:
                    if adj_matrix[state,p_state] == 0:
                    # print(f"Connecting {state} to {p_state} -> {next_state} ({possible_states}) {possible_actions}")
                        if adj_matrix[p_state,p_state] == 0:
                            adj_matrix[p_state,p_state] = 1
                        adj_matrix[state,p_state] = 1
                        adj_matrix[p_state,state] = 1

                # Linha onde ocorre a retencao do aprendizado
                q_table[state, action] = q_table[state, action] + alpha*(reward + gamma * np.max(q_table[next_state]) - q_table[state, action])
                state = next_state
                steps += 1

            sum_acumulated_reward.append(acumulated_reward)
            if idx == 1:
                print(f"{acumulated_reward}",end=" >> ")
                sys.stdout.flush()
            if idx == num_iter:
                print(f"{np.mean(sum_acumulated_reward,dtype=int)}",end=" | ")
                sys.stdout.flush()
        return q_table

    def choose_action(self, e, epsilon, state, table, adj_matrix, path_actions):
        possible_actions = e.action_mask(state)
        possible_states = e.actions_to_states(possible_actions)
        _, best_actions = e.sensor()
        if sum(best_actions) > 0:   # Usado quando o sensor encontra uma sujeira
            q_actions = np.multiply(table[state],best_actions)
            q_actions = np.add(q_actions,best_actions)
            action = np.argmax(q_actions)
        else:                       
            for ia, act in enumerate(possible_actions): # Verificar disponibilidade de acoes
                if act == 1:
                    s,_,_ = e.look(ia)
                    if adj_matrix[s,s] > 1 and ia != 4:
                        possible_actions[ia] = 0
                
            if sum(possible_actions) == 0 or len(path_actions) > 0: # Usado quando aspirador esta perdido
                if len(path_actions) == 0 or path_actions is None:
                    path = self.shortest_path(adj_matrix,state,table)
                    if path is not None:
                        path_actions = e.states_to_actions(path)
                    if path_actions is None or path is None:
                        print(
                            f"\nCurrent: {state} > ({path}) ({path_actions}) \nPossible States:\
                            \n{[(x,i) for x,i in enumerate(adj_matrix[state]) if x in possible_states]}\
                            \nMap Render:\
                            \n{e.render(adj_matrix)}"
                        )
                        sys.exit()
                        
                action = path_actions.pop(0) # Retornar para areas inexploradas
            elif np.random.uniform(0, 1) < epsilon:
                action = self.monte_carlo_selection(possible_actions)  # Explorar diferentes acoes
            else:
                actions = np.multiply(table[state],possible_actions)
                action = np.argmax(actions) # Utilizar valores aprendidos
        return action, possible_states, path_actions

    def sarsa(self, s_table, epsilon, e, num_steps=50, num_iter=500):
        alpha = 0.1
        gamma = 0.9 
        sum_acumulated_reward = []
        size = e.num_states
        for idx in range(1,num_iter+1):
            adj_matrix = np.zeros((size, size), dtype=int) # Matrix de adjacencia do grafo de exploracao  
            state, _ = e.reset()    # Resetando aspirador para um estado aleatorio
            acumulated_reward = 0

            adj_matrix[state,state] = 2 # Este valor indica na matrix que este estado ja foi explorado
            path_actions = []
            
            action,possible_states,path_actions = self.choose_action(e,epsilon,state,s_table,adj_matrix,path_actions)
            steps = 0
            while steps < num_steps:
                next_state, reward, _ = e.step(action)
                acumulated_reward += reward
                
                # Ajustes na matrix de adjacencia
                # Inserindo ligacoes entre estados (valor 1) e estados esplorados (valor 2)
                adj_matrix[next_state,next_state] = 2
                for p_state in possible_states:
                    if adj_matrix[state,p_state] == 0:
                    # print(f"Connecting {state} to {p_state} -> {next_state} ({possible_states}) {possible_actions}")
                        if adj_matrix[p_state,p_state] == 0:
                            adj_matrix[p_state,p_state] = 1
                        adj_matrix[state,p_state] = 1
                        adj_matrix[p_state,state] = 1
                        
                next_action, possible_states, path_actions = self.choose_action(e,epsilon,next_state,s_table,adj_matrix,path_actions)

                # Linha onde ocorre a retencao do aprendizado
                
                s_table[state, action] = s_table[state, action]+ alpha*(reward + gamma * s_table[next_state][next_action] - s_table[state, action])
                
                state = next_state
                action = next_action
                steps += 1

            sum_acumulated_reward.append(acumulated_reward)
            if idx == 1:
                print(f"{acumulated_reward}",end=" >> ")
                sys.stdout.flush()
            if idx == num_iter:
                print(f"{np.mean(sum_acumulated_reward,dtype=int)}",end=" | ")
                sys.stdout.flush()
        return s_table

    def learn(self, method_index=0):
        print(f'  Average Reward of\n  First and Last Iteration')
        table = np.zeros([self.map_size**2, self.model().num_actions])
        method = self.methods[method_index]
        for i in range(1, self.num_pop+1):
            print(f'Map {i}: ',end='')
            map = self.random_map(self.map_size, self.num_targets)
            env = self.model(map)
            table = method(table, self.epsilon, env, self.num_step, self.num_iter) 
            print('')

        self.policy = table
        
    def exec(self, env, state, num_steps=1, show=False):
        if self.policy is None:
            self.policy = np.zeros([self.map_size**2, self.model().num_actions])
            
        if self.adj_matrix is None: 
            size = env.num_rows*env.num_cols
            self.adj_matrix = np.zeros((size, size), dtype=int) 
        
        env.set_state(state)
        #state, _ = env.encode(x,y)
        reward = 0
        acumulated_reward = 0
        actions_taken = []
        states_arrived = []
        
        self.adj_matrix[state,state] = 2
        possible_actions = []
        path_actions = []
        frames = []
        for _ in range(num_steps):
            flag = 'Best'     
            possible_actions = env.action_mask(state)
            possible_states = env.actions_to_states(possible_actions)
            _, best_actions = env.sensor()
            if sum(best_actions) > 0:
                actions = np.multiply(self.policy[state],best_actions)
                actions = np.add(actions,best_actions)
                action = np.argmax(actions)
            else:
                for ia, act in enumerate(possible_actions):
                    if act == 1:
                        s,_,_ = env.look(ia)
                        if self.adj_matrix[s,s] > 1 and ia != 4:
                            possible_actions[ia] = 0
                actions = np.multiply(self.policy[state],possible_actions)
                if sum(actions) == 0 and sum(possible_actions) > 0:
                    actions = possible_actions
                if sum(actions) > 0 and len(path_actions) == 0:
                    flag = 'Policy'
                    action = np.argmax(actions)
                else:
                    flag = 'Return'
                    if len(path_actions) == 0:
                        path = self.shortest_path(self.adj_matrix,state,self.policy)
                        if path is not None:
                            path_actions = env.states_to_actions(path)
                        if path_actions is None or path is None:
                            print(
                                f"\nCurrent: {state} > ({path}) ({path_actions})\
                                \nAvailable Actions: {actions} = {self.policy[state]}*{possible_actions}\
                                \nPossible States:\
                                \n{[(x,i) for x,i in enumerate(self.adj_matrix[state]) if x in possible_states]}\
                                \nMap Render:\
                                \n{env.render(self.adj_matrix)}"
                            )
                            sys.exit()
                    action = path_actions.pop(0)

            frames.append({
                'frame': env.render(self.adj_matrix),
                'state': state,
                'action': env.last_action,
                'action_flag': flag,
                'best_actions': best_actions,
                'possible_actions': possible_actions,
                'reward': reward,
                'acumulated_reward': acumulated_reward,
            })

            actions_taken.append(action)
            next_state, reward, _ = env.step(action)
            states_arrived.append(next_state)

            self.adj_matrix[next_state,next_state] = 2
            for p_state in possible_states:
                if self.adj_matrix[state,p_state] == 0:
                    if self.adj_matrix[p_state,p_state] == 0:
                        self.adj_matrix[p_state,p_state] = 1
                    self.adj_matrix[state,p_state] = 1
                    self.adj_matrix[p_state,state] = 1

            state = next_state
            acumulated_reward += reward
        
        if show:
            self.show_frames(frames)
        return actions_taken, states_arrived
        
        

        self.policy


class utils:
    def merge_dicts(dict1: dict, dict2: dict) -> dict:
        for key, value in dict1.items():
            if key in dict2 and isinstance(value, dict):
                for inner_key, inner_value in value.items():
                    if inner_key in dict2[key] and isinstance(inner_value, set):
                        dict2[key][inner_key].update(inner_value)
                    else:
                        dict2[key][inner_key] = inner_value 
            else:
                dict2[key] = value
        return dict2
    
    def set_changes(original:set, changes:set):
        intersection = original.intersection(changes)
        added = changes - original
        removed = original - intersection
        return intersection.union(added), added, removed


import threading
from dataclasses import dataclass, field
from maspy.environment import Environment
from maspy.communication import Channel
from maspy.error import (
    InvalidBeliefError,
    InvalidPlanError,
    RunPlanError,
)
from maspy.utils import utils
from typing import TypeVar, List, Optional, Dict, Set, Tuple, Any
from collections.abc import Iterable, Callable
from time import sleep
import importlib as implib
import inspect

gain = TypeVar('gain')
lose = TypeVar('lose')
test = TypeVar('test')

DEFAULT_SOURCE = "self"

@dataclass(eq=True, frozen=True)
class Belief:
    key: str
    _args: tuple = field(default_factory=tuple)
    source: str = DEFAULT_SOURCE

    @property
    def args(self):
        if len(self._args) > 1:
            return self._args
        elif len(self._args) == 1:
            return self._args[0]
        else:
            return tuple()
        
    @property
    def args_len(self):
        return len(self._args)

    def __post_init__(self):
        match self._args:
            case list() | dict() | str():
                object.__setattr__(self, "_args", tuple([self._args]))
            case tuple():
                pass
            case Iterable():
                object.__setattr__(self, "_args", tuple(self._args))
            case _:
                object.__setattr__(self, "_args", tuple([self._args]))

    def weak_eq(self, other: "Belief"):
        return (
            self.key == other.key
            and len(self._args) == len(other._args)
            and self.source == other.source
        )

    def update(self, key: str = None, args=None, source=None) -> "Belief":
        if key is not None:
            new_name = key
        else:
            new_name = self.key

        if args is not None:
            new_args = args
        else:
            new_args = self._args

        if source is not None:
            new_source = source
        else:
            new_source = self.source

        return Belief(new_name, new_args, new_source)

    # implementing hash for Belief is fine, it is impossible to change something inside
    # without creating a new object therefore, Belief can be used in dict and sets
    def __hash__(self) -> int:
        args_hashable = []
        unhashable_types = {}
        for arg in self._args:
            arg_dict = type(arg).__dict__
            if arg_dict.get("__hash__"):
                args_hashable.append(arg)
            elif isinstance(arg, (List, Dict, Set)):
                args_hashable.append(repr(arg))
            else:
                raise TypeError(f"Unhashable type: {type(arg)}")
        args_hashable = tuple(args_hashable)

        return hash((self.key, args_hashable, self.source))
    
    def __str__(self) -> str:
        return f"Belief{self.key,self.args,self.source}"

@dataclass
class Goal:
    key: str
    _args: tuple = field(default_factory=tuple)
    source: str = DEFAULT_SOURCE

    @property
    def args(self):
        if len(self._args) > 1:
            return self._args
        elif len(self._args) == 1:
            return self._args[0]
        else:
            return tuple()
        
    @property
    def args_len(self):
        return len(self._args)

    def __post_init__(self):
        match self._args:
            case list() | dict() | str():
                object.__setattr__(self, "_args", tuple([self._args]))
            case tuple():
                pass
            case Iterable():
                object.__setattr__(self, "_args", tuple(self._args))
            case _:
                object.__setattr__(self, "_args", tuple([self._args]))

    def weak_eq(self, other: "Goal"):
        return (
            self.key == other.key
            and len(self._args) == len(other._args)
            and self.source == other.source
        )

    def update(self, key: str = None, args=None, source=None) -> "Goal":
        if key is not None:
            new_name = key
        else:
            new_name = self.key

        if args is not None:
            new_args = args
        else:
            new_args = self._args

        if source is not None:
            new_source = source
        else:
            new_source = self.source

        return Goal(new_name, new_args, new_source)

    def __hash__(self) -> int:
        args_hashable = []
        unhashable_types = {}
        for arg in self._args:
            arg_dict = type(arg).__dict__
            if arg_dict.get("__hash__"):
                args_hashable.append(arg)
            elif isinstance(arg, (List, Dict, Set)):
                args_hashable.append(repr(arg))
            else:
                raise TypeError(f"Unhashable type: {type(arg)}")
        args_hashable = tuple(args_hashable)

        return hash((self.key, args_hashable, self.source))
    
    def __str__(self) -> str:
        return f"Goal{self.key,self.args,self.source}"
    
@dataclass
class Event:
    change: str # ( + or add, - or rm, ~ or test )
    data: Belief | Goal 
    intention: str = None # still don't understand how it works
    
    def __init__(self,change,data,intention=None):
        if type(change)==TypeVar: self.change = change.__name__ 
        else: self.change = change
        self.data = data
        self.intention = intention
    
    def __str__(self) -> str:
        return f"Event{self.change,self.data,self.intention}"
    
@dataclass
class Plan:
    trigger: Event
    context: list = field(default_factory=list)
    body: Callable = None
    
    def __str__(self) -> str:
        return f"Plan{self.trigger,self.context,self.body.__name__}"
    
@dataclass
class Ask:
    data_type: Belief | Goal | str
    reply: list = field(default_factory=list)
    source: str = "unknown"

MSG = Belief | Ask | Goal | Plan

_type_env_set = {Environment, "environment", "envrmnt", "env"}
_type_ch_set = {Channel, "channel", "chnnl", "ch", "c"}

def pl(change, data: Belief | Goal, context: Belief | Goal | Iterable = []):
    class decorator:
        def __init__(self,func):
            self.func = func
                
        def __set_name__(self, instance: Agent, name: str):
            if not isinstance(data, Belief | Goal):
                raise TypeError
            if isinstance(context, Belief | Goal):
                list_context = [context]
            if isinstance(context, Iterable):
                list_context = context
                for ctxt in list_context:
                    if isinstance(ctxt, Belief | Goal):
                        continue
                    else:
                        raise Exception(f'Invalid type {type(ctxt)}:{ctxt} - was expecting Belief or Goal')
            
            event = Event(change,data)
            plan = Plan(event,list_context,self.func)
            try:
                instance._plans += [plan]
            except AttributeError:
                instance._plans = [plan]
    return decorator

class Agent:
    def __init__(
        self,
        name: str,
        beliefs: Optional[Iterable[Belief] | Belief] = None,
        goals: Optional[Iterable[Goal] | Goal] = None,
        full_log = False,
        log_type = "Default",
        instant_mail = False
    ):              
        self.stop_flag = None
        self.running = False
        self.thread = None
        self.saved_msgs = []
        self.full_log = full_log
        self.log_type = log_type
        
        self.my_name = name
        from maspy.admin import Admin
        Admin().add_agents(self)
        self._name = f"Agent:{self.my_name}"
        
        self._environments: Dict[str, Environment] = dict()
        self._channels: Dict[str, Channel] = dict()

        self.__events: List[Event] = []
        self.__beliefs: Dict[str, Dict[str, Set[Belief]]] = dict()
        self.__goals: Dict[str, Dict[str, Set[Goal]]] = dict()
        if beliefs: self.add(beliefs)
        if goals: self.add(goals)
        
        self.instant_mail = instant_mail
        self.connect_to(Channel(),"")
        self.paused_agent = False
        
        try: 
            self._plans
        except AttributeError:
            self._plans = []
         
        #self.print(f"Initialized") 

    def start(self):
        self.reasoning()
    
    def print(self,*args, **kwargs):
        return print(f"{self._name}>",*args,**kwargs)
    
    @property
    def print_beliefs(self):
        print("Beliefs:",self.__beliefs)

    @property
    def print_goals(self):
        print("Goals:",self.__goals)
    
    @property
    def print_plans(self):
        print("Plans:",self._plans)
    
    @property
    def print_events(self):
        print("Events:",self.__events)
    
    def connect_to(self, target: Channel | Environment | str, target_name: str):
        match target:
            case Environment():
                self._environments[target._my_name] = target
            case Channel():
                self._channels[target._my_name] = target
            case str():
                classes = []
                try:
                    imported = implib.import_module(target)
                except ModuleNotFoundError:
                    self.print(f"No File named '{target_name}'")
                    return
                for name, obj in inspect.getmembers(imported):
                    if inspect.isclass(obj) and name != "Environment" and name != "Channel":
                        lineno = inspect.getsourcelines(obj)[1]
                        classes.append((lineno, obj))
                    if name == "Environment":
                        connect_list = self._environments
                    if name == "Channel":
                        connect_list = self._channels
                classes.sort()
                target = classes[0][1](target_name)
                connect_list[target_name] = target
                del imported
                
        target.add_agents(self)
        return target

    def disconnect_from(self, target: Channel | Environment):
        match target:
            case Environment():
                del self._environments[target._my_name]
            case Channel():
                del self._channels[target._my_name]
                
    def get_env(self, env_name: str):
        return self._environments[env_name]
    
    def add_plan(self, plan: List[Plan | Tuple] | Plan | Tuple):
        plans = self._clean_plans(plan)
        self._plans += plans

    def rm_plan(self, plan: Plan):
        self._plans.pop(plan)
    
    def _new_event(self,change,data,intention=None):
        try:
            for dt in data:
                self.print(f"New Event: {change}:{dt}")  if self.full_log else ...
                self.__events.append(Event(change,dt,intention))
        except(TypeError):
            self.print(f"New Event: {change}:{data}")  if self.full_log else ...
            self.__events.append(Event(change,data,intention))
    
    def _get_type_base(self, data_type):
        if data_type == Belief:
            return self.__beliefs
        elif data_type == Goal:
            return self.__goals
        elif data_type == Plan:
            return self._plans
        elif data_type == Event:
            return self.__events
        else:
            print(f"Type is neither Belief | Goal | Plan | Event : {data_type}")
            return None
    
    def add(self, data_type: Belief | Goal | Iterable[Belief | Goal]):
        self.print(f"Adding {data_type}") if self.full_log else ...
        cleaned_data = self._clean(data_type)
        
        for type_data, data in cleaned_data.items():
            if len(data) == 0: continue
            type_base = self._get_type_base(type_data)
            type_base = utils.merge_dicts(data,type_base)
            
        self._new_event("gain",data_type)
                 
    def rm(self, data_type: Belief | Goal | Iterable[Belief | Goal]):
        self.print(f"Removing {data_type}") if self.full_log else ...
        
        if type(data_type) != Iterable: data_type = [data_type]
            
        for typ in data_type:
            type_base = self._get_type_base(type(typ))
            type_base[typ.source][typ.key].remove(typ)
        
        self._new_event("lose",data_type)

    def get(self, data_type: Belief | Goal | Plan | Event,
        search_with:  Belief | Goal | Plan | Event = None,
        all = False, ck_chng=True, ck_type=True, ck_args=True, ck_src=True
    ) -> List[Belief | Goal | Plan | Event]:
        if type(data_type) is type: data_type = data_type(0,0,0)
        type_base = self._get_type_base(type(data_type))
        if search_with is None: search_with = data_type

        change, data = self._to_belief_goal(search_with)
        
        found_data = []
        match data_type:
            case Belief() | Goal(): 
                for keys in type_base.values():
                    for values in keys.values():
                        for value in values:
                            if self._compare_data(value,data,ck_type,ck_args,ck_src):
                                found_data.append(value)
                                if not all: return found_data
                                
            case Plan() | Event(): 
                for plan_event in type_base:
                    chng, belf_goal = self._to_belief_goal(plan_event)
                    
                    if change and ck_chng and chng != change:
                        continue
                    if self._compare_data(belf_goal,data,ck_type,ck_args,ck_src):
                        found_data.append(plan_event)
                        if not all: return found_data

            case _: pass
        return found_data

    def _to_belief_goal(self, data_type: Belief | Goal | Plan | Event):
        change = None
        belief_goal:Belief | Goal = None
        match data_type:
            case Belief() | Goal():
                belief_goal = data_type
            case Plan(): 
                change: str = data_type.trigger.change
                belief_goal = data_type.trigger.data
            case Event(): 
                change = data_type.change
                belief_goal = data_type.data
            case _: 
                self.print(f"Error in _to_belief_goal: {type(data_type)}:{data_type}")
                return None, None
        return change,belief_goal
    
    def _compare_data(self, data1, data2, ck_type,ck_args,ck_src):
        #self.print(f"Comparing: \n\t{data1} and {data2}")
        if ck_type and type(data1) != type(data2):
            #self.print("Failed at type")
            return False
        if data1.key != data2.key:
            #self.print("Failed at key")
            return False
        if ck_src and data2.source != DEFAULT_SOURCE and data1.source != data2.source:
            #self.print("Failed at source")
            return False
        if not ck_args:
            return True
        if data1.args_len != data2.args_len:
            #self.print("Failed at args_len")
            return False
        for arg1,arg2 in zip(data1._args,data2._args):
            if isinstance(arg1,str) and (arg1[0].isupper()):
                continue
            elif isinstance(arg2,str) and (arg2[0].isupper()):
                continue
            elif arg1 == arg2:
                continue
            else:
                #self.print(f"Failed at args {arg1} x {arg2}")
                return False
        else:
            #self.print("Data is Compatible")
            return True
                      
    def _run_plan(self, plan: Plan, trigger: Belief | Goal, args: tuple):
        self.print(f"Running {plan}")  if self.full_log else ...
        try:
            #self.print(f'running with {trigger._args} {args}')
            return plan.body(self, trigger.source, *trigger._args, *args)
        except KeyError:
            self.print(f"{plan} doesn't exist")
            raise RunPlanError

    # TODO: implement stoping plan
    def _stop_plan(self, plan):
        self.print(f"Stoping {plan})")  if self.full_log else ...
        pass
    
    def send(self, target: str | tuple | List, act: TypeVar, msg: MSG | str, channel: str = Channel()._my_name):  
        try:
            self._channels[channel]._send(self.my_name,target,act,msg)
        except KeyError:
            self.print(f"Not Connected to Selected Channel:{channel}")
    
    def save_msg(self, act, msg):
        if self.instant_mail: 
            self.recieve_msg(act,msg)
        else:
            self.saved_msgs.append((act,msg))

    def _mail(self, selection_function="ALL"):
        if selection_function not in {None,"ALL"}:
            selection_function(self.saved_msgs)
        else:
            while self.saved_msgs:
                act,msg = self.saved_msgs.pop()
                self.recieve_msg(act,msg)
                if selection_function != "ALL":
                    break

    def recieve_msg(self, act, msg: MSG):
        #self.print(f"Received from {sender} : {act} -> {msg}")  if self.full_log else ...
        match act:
            case "tell" | "achieve":
                self.add(msg)

            case "untell" | "unachieve":
                self.rm(msg)

            case "askOne":
                found_data = self.get(msg.data_type,ck_src=False)
                self.send(msg.source, TypeVar('tell'), found_data)

            case "askAll":
                found_data = self.get(msg.data_type,all=True,ck_src=False)
                for data in found_data:
                    self.send(msg.source, TypeVar('tell'), data)

            case "tellHow":
                self.add_plan(msg)

            case "untellHow":
                self.rm_plan(msg)

            case "askHow":
                found_plans = self.get(Plan,msg.data_type,all=True)
                for plan in found_plans:
                    self.send(msg.source, TypeVar('tellHow'), plan)
            case _:
                TypeError(f"Unknown type of message {act}:{msg}")
    
    def find_in(self, agent_name, cls_type=None, cls_name=["env","default"], cls_instance=None):
        cls_type = cls_type.lower()
        try:
            if cls_instance:
                return cls_instance.agent_list[agent_name]
            if cls_type in _type_env_set:
                cls_name = cls_name[0] if type(cls_name) == list else cls_name  
                return self._environments[cls_name].agent_list[agent_name]
            if cls_type in _type_ch_set:
                cls_name = cls_name[1] if type(cls_name) == list else cls_name  
                return self._channels[cls_name].agent_list[agent_name]
        except KeyError as ke:
            self.print(f"Not connected to {cls_type}:{cls_name}:{ke}")
            
    def execute_in(self,env_name) -> Environment:
        try:
            return self._environments[env_name]
        except KeyError:
            self.print(f"Not Connected to Environment:{env_name}")

    def send_msg(self, target: str, act: str, msg: MSG, channel: str):
        pass

    def reasoning(self):
        #self.print(f"Starting Reasoning")
        self.running = True
        self.stop_flag = threading.Event()
        self.thread = threading.Thread(target=self.cycle,args=(self.stop_flag,))
        self.thread.start()
    
    def stop_cycle(self):
        self.print("Shutting Down...")
        self.stop_flag.set()
        self.paused_agent = True
        self.running = False
            
    def cycle(self, stop_flag):
        while not stop_flag.is_set():   
            self._perception()   
            self._mail()  
            event = self._select_event()
            #self.print(f"Selected event: {event} in {self.__events}") 
            plans = self._retrieve_plans(event)
            #self.print(f"Selected plans: {plans} in {self._plans}")
            chosen_plan, args = self._select_plan(plans,event)
            #self.print(f"Selected chosen_plan: {chosen_plan} with {args} arguments")
            result = self._execute_plan(chosen_plan,event,args)
            #sleep(1)

    def _perception(self):
        percept_dict = dict()
        for env_name in self._environments:
            self.print(f"Percepting '{env_name}'") if self.full_log else ...
            percept_dict.update(self._environments[env_name].perception())
        percept_dict = self._percepts_to_beliefs(percept_dict)
        self._revise_beliefs(percept_dict)
    
    def _percepts_to_beliefs(self,percepts: dict) -> dict:
        # TODO make use of percepts group
        for source, keys in percepts.copy().items():
            for key,percepts_set in keys.items():
                belief_set = set()
                for percept in percepts_set:
                    belief_set.add(Belief(percept.key,percept.args,source))
                percepts[source][key] = belief_set
        return percepts
     
    def _revise_beliefs(self, percept_dict: dict):
        for source, keys in self.__beliefs.copy().items():
            if source == DEFAULT_SOURCE: continue
            if source in percept_dict:
                for key, beliefs in keys.copy().items():
                    if key in percept_dict[source]: 
                        new, gain, lose = utils.set_changes(beliefs,percept_dict[source][key])
                        self.__beliefs[source][key] = new
                        self._new_event("gain",gain) # Gained new specific belief
                        self._new_event("lose",lose) # Lost an old specific belief
                        del percept_dict[source][key]
                    else:
                        self._new_event("lose",self.__beliefs[source][key]) # Lost whole key belief
                        del self.__beliefs[source][key]
                        
                if percept_dict[source] == {}:
                    del percept_dict[source]
            else:
                for beliefs in keys.values():
                    self._new_event("lose",beliefs) # Lost whole source of belief (env)
                del self.__beliefs[source]
        
        for keys in percept_dict.values():
            for beliefs in keys.values():
                self._new_event("gain",beliefs) # Gained beliefs of new sources/keys
        self.__beliefs.update(percept_dict)
    
    def _select_event(self):
        if self.__events == []: return None
        return self.__events.pop(0)
    
    def _retrieve_plans(self, event):
        if event is None: return None
        return self.get(Plan,event,all=True,ck_src=False)
    
    def _select_plan(self, plans, event:Event):
        if plans is None: return None, None
        
        args = tuple()
        for plan in plans:
            for context in plan.context:
                ctxt = self.get(context,ck_src=False)
                if not ctxt:
                    break
                for x in ctxt[0]._args:
                    args += (x,)
            else:    
                return plan, args
        self.print(f"Found no applicable plan for {event.change}:{event.data}")
        return None, None
    
    def _execute_plan(self, chosen_plan:Plan, event: Event, args):
        if not chosen_plan:
            return None
        try:
            return self._run_plan(chosen_plan,event.data,args)
        except RunPlanError:
            self.print(f"{chosen_plan} failed")

    # TODO: should invalid arguments be an error or a warning?
    def _clean(
        self, data_type: Iterable[Belief | Goal] | Belief | Goal 
    ) -> Dict:
        type_dicts = {Belief: dict(), Goal: dict()}
        match data_type:
            case None:
                pass
            case Belief() | Goal():
                type_dicts[type(data_type)].update({data_type.source: {data_type.key: {data_type}}})
            case Iterable():
                for typ in data_type:
                    if not isinstance(typ, Belief) and not isinstance(typ, Goal):
                        raise InvalidBeliefError(
                            f"Expected data type to be Iterable[Belief | Goal] | Belief | Goal, recieved Iterable[{type(typ).__name__}]"
                        )

                    type_dict = type_dicts[type(typ)]
                    if typ.source in type_dict:
                        if typ.key in type_dict[typ.source]:
                            type_dict[typ.source][typ.key].add(typ)
                        else:
                            type_dict[typ.source].update({typ.key: {typ}})
                    else:
                        type_dict.update({typ.source: {typ.key: {typ}}})
            case _:
                raise InvalidBeliefError(
                    f"Expected data type to have be Iterable[Belief | Goal] | Belief | Goal, recieved {type(data_type).__name__}"
                )    
        return type_dicts
    
    def _clean_plans(
        self,
        plans: Optional[Iterable[Plan] | Plan ],
    ) -> List[Plan]:
        match plans:
            case None:
                return []
            case Plan():
                return [plans]
            case Iterable():
                plan_list = []
                for plan in plans:
                    if isinstance(plan, Plan):
                        plan_list.append(plan)
                    if isinstance(plan,Tuple):
                        plan_list.append(Plan(*plan))
                        
                return plan_list
            case _:
                raise InvalidPlanError(
                    f"Expected plans to have type Dict[str, Callable] | Iterable[Tuple[str, Callable]] | Tuple(str, Callable), recieved {type(plans).__name__}"
                )




from threading import Lock
from collections.abc import Iterable
from typing import Dict, TypeVar

Agt_name = TypeVar('Agt_name')
Agt_cls = TypeVar('Agt_cls')
Agt_inst = TypeVar('Agt_inst')
Agt_fullname = TypeVar('Agt_fullname')

tell= TypeVar('tell')
untell = TypeVar('untell')
tellHow = TypeVar('tellHow')
untellHow = TypeVar('untellHow')
achieve = TypeVar('achieve')
unachieve = TypeVar('unachieve')
askOne = TypeVar('askOne')
askAll = TypeVar('askAll')
askHow = TypeVar('askHow')

broadcast = TypeVar('broadcast')

class CommsMultiton(type):
    _instances: Dict[str, "Channel"] = {}
    _lock: Lock = Lock()

    def __call__(cls, __my_name="default"):
        with cls._lock:
            if __my_name not in cls._instances:
                instance = super().__call__(__my_name)
                cls._instances[__my_name] = instance
        return cls._instances[__my_name]


class Channel(metaclass=CommsMultiton):
    def __init__(self, comm_name) -> None:
        from maspy.agent import Belief, Goal, Ask, Plan
        self.data_types = {Belief,Goal,Ask,Plan}
        self._my_name = comm_name
        self.agent_list: Dict[Agt_cls, Dict[Agt_name, Agt_fullname]] = {}
        self._agents: Dict[Agt_fullname, Agt_inst] = {}
        self._name = f"{type(self).__name__}:{self._my_name}"
        self.full_log = True
        
    def print(self,*args, **kwargs):
        return print(f"{self._name}>",*args,**kwargs)
    
    def add_agents(self, agents):
        try:
            for agent in agents:
                self._add_agent(agent)
        except TypeError:
            self._add_agent(agents)

    def _add_agent(self, agent):
        if type(agent).__name__ in self.agent_list:
            if agent.my_name[0] in self.agent_list[type(agent).__name__]:
                self.agent_list[type(agent).__name__][agent.my_name[0]].update({agent.my_name})
                self._agents[agent.my_name] = agent
            else:
                self.agent_list[type(agent).__name__].update({agent.my_name[0] : {agent.my_name}})
                self._agents[agent.my_name] = agent
        else:
            self.agent_list[type(agent).__name__] = {agent.my_name[0] : {agent.my_name}}
            self._agents[agent.my_name] = agent
        
        self.print(f'Connecting agent {type(agent).__name__}:{agent.my_name}')

            
    def _rm_agents(self, agents):
        try:
            for agent in agents:
                self._rm_agent(agent)
        except TypeError:
            self._rm_agent(agents)

    def _rm_agent(self, agent):
        if agent.my_name in self._agents:
            del self._agents[agent.my_name]
            del self.agent_list[type(agent).__name__][agent.my_name[0]]
        self.print(
            f"Desconnecting agent {type(agent).__name__}:{agent.my_name}"
        )

    def _send(self, sender, target, act, message):  
        if type(act) == TypeVar: act = act.__name__ 
        
        messages = []
        if type(message) == Iterable:
            for m in message:
                messages.append(self.parse_sent_msg(sender,act,m))
        else:
            messages.append(self.parse_sent_msg(sender,act,message))

        for msg in messages:
            if type(target) == Iterable:
                for trgt in target:
                    self._sending(sender,trgt,act,msg)
            elif target == broadcast:
                for agent_name in self._agents.keys():
                    self._sending(sender,agent_name,act,msg)
            else:
                self._sending(sender,target,act,msg)
    
    def _sending(self, sender, target, act, msg):
        self.print(f"{sender} sending {act}:{msg} to {target}") if self.full_log else ...        
    
        try:         
            self._agents[target].save_msg(act,msg)
        except KeyError:
            self.print(f"Agent {target} not connected")
    
    def parse_sent_msg(self,sender, act, msg):
        from maspy.agent import Belief, Goal, Ask, Plan
        if type(msg) in {Belief, Goal} and msg is not None:
            msg = msg.update(source=sender)
        if act in {"askOne","askAll","askHow"}:
            msg = Ask(msg, source=sender)
        return msg